# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Any, Callable, Dict, Optional, TypeAlias, Union

import dace
from dace import (
    properties as dace_properties,
    subsets as dace_subsets,
    transformation as dace_transformation,
)
from dace.sdfg import (
    graph as dace_graph,
    nodes as dace_nodes,
    propagation as dace_propagation,
    utils as dace_sdutils,
)
from dace.transformation.dataflow import map_fusion_helper as dace_mfhelper
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import (
    sdfg_args as gtx_dace_args,
    transformations as gtx_transformations,
)
from gt4py.next.program_processors.runners.dace.transformations import (
    map_fusion_utils as gtx_mfutils,
    splitting_tools as gtx_dace_split,
)


VerticalMapSplitCallback: TypeAlias = Callable[
    ["VerticalSplitMapRange", dace_nodes.MapExit, dace_nodes.MapEntry, dace.SDFGState, dace.SDFG],
    bool,
]
"""Callback used to influence the behaviour of `VerticalSplitMapRange`.

The function is used by `can_be_applied()`, before any other check is performed, thus
it might be that the split would not apply. If the function returns `True` then the
transformation will continue and if `False` is returned then the transformation will
not continue. This is similar to the callback function offered by `MapFusionVertical`
or `HorizontalMapSplitCallback`.

It has the following arguments:
- The transformation object itself.
- The MaxExit node of the first Map.
- The MapEntry node of the second Map.
- The SDFGState in which the nodes where found.
- The SDFG itself.
"""


HorizontalMapSplitCallback: TypeAlias = Callable[
    [
        "HorizontalSplitMapRange",
        dace_nodes.MapEntry,
        dace_nodes.MapEntry,
        dace.SDFGState,
        dace.SDFG,
    ],
    bool,
]
"""Callback used to influence the behaviour of `HorizontalSplitMapRange`.

The function is used by `can_be_applied()`, before any other check is performed, thus
it might be that the split would not apply. If the function returns `True` then the
transformation will continue and if `False` is returned then the transformation will
not continue. This is similar to the callback function offered by `MapFusionHorizontal`
or `VerticalMapSplitCallback`.

It has the following arguments:
- The transformation object itself.
- The MapEntry node of the first Map.
- The MapEntry node of the second Map.
- The SDFGState in which the nodes where found.
- The SDFG itself.
"""


def gt_horizontal_map_split_fusion(
    sdfg: dace.SDFG,
    run_simplify: bool,
    fuse_map_fragments: bool,
    consolidate_edges_only_if_not_extending: bool,
    skip: Optional[set[str]] = None,
    run_map_fusion: bool = False,
    check_split_callback: Optional[HorizontalMapSplitCallback] = None,
    check_fusion_callback: Optional["gtx_transformations.HorizontalMapFusionCallback"] = None,
    only_if_common_ancestor: bool = True,
    validate: bool = True,
    validate_all: bool = False,
) -> int:
    """Performs horizontal map splitting on the provided SDFG.

    This function runs `HorizontalSplitMapRange` transformation until a fix point is
    reached. Due to a limitation that transformation might not fuse all possible
    Maps together. For this case the `run_map_fusion` flag is provided.
    If it is set, then `MapFusionHorizontal` is also called. It is important to
    note that Map fusion operates on all Maps in the SDFG.

    Args:
        sdfg: The SDFG on which we operate.
        run_simplify: Run `gt_simplify()` at the end.
        fuse_map_fragments: Directly fuse the Maps inside `HorizontalSplitMapRange`.
        consolidate_edges_only_if_not_extending: See `MapFusionHorizontal` for more.
        skip: Skip these transformation during simplification.
        run_map_fusion: If `True` also run `MapFusionHorizontal`. Note
            that this will fuse Maps globally.
        check_split_callback: Use this as callback for the split transformation, see
            `HorizontalMapSplitCallback` for more information.
        check_fusion_callback: Use this as callback for the `MapFusionHorizontal`
            transformation, it only has an effect if `run_map_fusion`
            has been set to `True`. Note that this callback is not passed to
            `HorizontalSplitMapRange` thus it can not be used to limit fusion of Maps
            that are created during the splitting, see `fuse_map_fragments`.
        only_if_common_ancestor: Passed to `MapFusionHorizontal` only has an effect
            if `run_map_fusion` is `True`.
        validate: Perform validation during the steps.
        validate_all: Perform extensive validation.
    """

    transformations = [
        HorizontalSplitMapRange(
            fuse_map_fragments=fuse_map_fragments,
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
            check_split_callback=check_split_callback,
        )
    ]

    if run_map_fusion:
        transformations.append(
            gtx_transformations.MapFusionHorizontal(
                check_fusion_callback=check_fusion_callback,
                only_toplevel_maps=True,
                consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
                only_if_common_ancestor=only_if_common_ancestor,
            )
        )

    ret = sdfg.apply_transformations_repeated(
        transformations,
        validate=False,
        validate_all=validate_all,
    )

    if run_simplify:
        if skip is None:
            skip = gtx_transformations.constants.GT_SIMPLIFY_DEFAULT_SKIP_SET
        if not consolidate_edges_only_if_not_extending:
            skip = skip.union(["ConsolidateEdges"])
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=False,
            validate_all=validate_all,
            skip=skip,
        )
    elif validate and (not validate_all):
        sdfg.validate()

    return ret


def gt_vertical_map_split_fusion(
    sdfg: dace.SDFG,
    run_simplify: bool,
    run_map_fusion: bool,
    consolidate_edges_only_if_not_extending: bool,
    fuse_map_fragments: bool,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
    skip: Optional[set[str]] = None,
    check_split_callback: Optional[VerticalMapSplitCallback] = None,
    check_fusion_callback: Optional["gtx_transformations.VerticalMapFusionCallback"] = None,
    validate: bool = True,
    validate_all: bool = False,
) -> int:
    """Performs vertical map splitting on the provided SDFG.

    The function essentially runs `VerticalSplitMapRange` until a fix point is reached.

    Note that if the function is called with `fuse_map_fragments` set to `True`,
    then the `VerticalSplitMapRange` will also run `MapFusionVertical` and
    `SplitAccessNode` but their activities are restricted to Maps that were created
    by the splitting.
    This is a different behaviour compared to `run_map_fusion` where `MapFusionVertical`
    is run as an independent transformation, which acts on the entire SDFG, it will
    however, respect `check_fusion_callback`.

    Args:
        sdfg: The SDFG on which we operate.
        run_simplify: Run `gt_simplify()` at the end.
        consolidate_edges_only_if_not_extending: See `MapFusionVertical` for more.
        fuse_map_fragments: Run `MapFusionVertical` directly inside the split transformation.
            Note that this is limited to the Maps that are were created. Furthermore,
            `check_fusion_callback` is not used.
        single_use_data: Precomputed single use data.
        skip: Skip these transformation during simplification.
        run_map_fusion: Also run `MapFusionVertical`. Note that it acts on the entire SDFG.
            This call uses `check_fusion_callback`.
        check_split_callback: Use this as callback for the split transformation, see
            `VerticalMapSplitCallback` for more information.
        check_fusion_callback: Use this as callback for `MapFusionVertical`.
        validate: Perform validation during the steps.
        validate_all: Perform extensive validation.

    Note:
        - In previous versions this function also called `MapFusionVertical` and
            `SplitAccessNode` without any restriction, which had the side effect
            that all Maps in the SDFG were subject to Map fusion.
        - Due to a bug in the transformation, not all Maps, that were created by
            the splitting were fused. Especially "chains" might still be present.
    """
    if single_use_data is None:
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

    transformations = [
        VerticalSplitMapRange(
            only_toplevel_maps=True,
            check_split_callback=check_split_callback,
            consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
            fuse_map_fragments=fuse_map_fragments,
            single_use_data=single_use_data,
        )
    ]

    if run_map_fusion:
        transformations.append(
            gtx_transformations.MapFusionVertical(
                only_toplevel_maps=True,
                check_fusion_callback=check_fusion_callback,
                consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
            )
        )
        transformations[-1]._single_use_data = single_use_data

    ret = sdfg.apply_transformations_repeated(
        transformations,
        validate=False,
        validate_all=validate_all,
    )

    if run_simplify:
        if skip is None:
            skip = gtx_transformations.constants.GT_SIMPLIFY_DEFAULT_SKIP_SET
        if not consolidate_edges_only_if_not_extending:
            skip = skip.union(["ConsolidateEdges"])
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=validate,
            validate_all=validate_all,
            skip=skip,
        )
    elif validate and (not validate_all):
        sdfg.validate()

    return ret


@dace_properties.make_properties
class SplitMapRange(dace_transformation.SingleStateTransformation):
    """
    Identify overlapping range between maps, and split the maps into
    maps with smaller ranges in order to create two maps with common
    range that can be fused together using the `HorizontalSplitMapRange`
    or `VerticalSplitMapRange` transformations.
    For example, given a map with range [0:10, 14:80] and a second map
    with range [5:15, 0:80], this transformation will split the first map
    into two maps with ranges [0:5, 14:80] and [5:10, 14:80], and the second
    map into four maps with ranges [0:10, 0:14], [0:10, 14:80] [10:15, 0:14]
    and [10:15, 14:80]. This will allow the `HorizontalSplitMapRange` or
    `VerticalSplitMapRange` transformations to fuse the maps together.
    """

    only_toplevel_maps = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Only perform fusing if the Maps are in the top level.",
    )

    def __init__(
        self,
        only_toplevel_maps: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if only_toplevel_maps is not None:
            self.only_toplevel_maps = only_toplevel_maps

    def split_maps(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
        first_map_entry: dace_nodes.MapEntry,
        first_map_exit: dace_nodes.MapExit,
        second_map_entry: dace_nodes.MapEntry,
        second_map_exit: dace_nodes.MapExit,
    ) -> tuple[list[dace_nodes.MapEntry], list[dace_nodes.MapEntry]]:
        """Split the map range in order to obtain an overlapping range between the first and second map.

        The function returns a pair of lists. The first `list` contains the new `MapEntry`s that
        were generated by splitting the first Map and the second `list` contains the `MapEntry`s
        generated by splitting the second Map.
        """
        splitted_range = gtx_mfutils.split_overlapping_map_range(
            first_map_entry.map, second_map_entry.map
        )
        assert splitted_range is not None

        scope_dict = graph.scope_dict()
        assert scope_dict[first_map_entry] is scope_dict[second_map_entry]
        containing_scope_or_none = scope_dict[first_map_entry]

        first_map_splitted_range, second_map_splitted_range = splitted_range

        def _replace_ranged_map(
            map_entry: dace_nodes.MapEntry,
            map_exit: dace_nodes.MapExit,
            new_ranges: list[dace_subsets.Range],
        ) -> list[dace_nodes.MapEntry]:
            """Replace a map with multiple maps based on new_ranges."""
            new_map_entries: list[dace_nodes.MapEntry] = []
            for i, r in enumerate(new_ranges):
                new_map_entry, _ = gtx_mfutils.copy_map_graph_with_new_range(
                    sdfg, graph, map_entry, map_exit, r, str(i)
                )
                new_map_entries.append(new_map_entry)
            gtx_mfutils.delete_map(graph, map_entry, map_exit)  # remove the original first map
            return new_map_entries

        new_map_entries_from_first = _replace_ranged_map(
            first_map_entry,
            first_map_exit,
            first_map_splitted_range,
        )
        new_map_entries_from_second = _replace_ranged_map(
            second_map_entry,
            second_map_exit,
            second_map_splitted_range,
        )

        # Now we propagate the Memlets. If the original Maps where located at the global
        #  scope we have to propagate the maps from the new ones. However, if they are
        #  not at global scope we simply propagate the surrounding Map, since they are
        #  by assumption inside the same Map.
        if containing_scope_or_none is None:
            for new_map_entry in new_map_entries_from_first + new_map_entries_from_second:
                dace_sdutils.canonicalize_memlet_trees_for_map(graph, new_map_entry)
                dace_propagation.propagate_memlets_map_scope(sdfg, graph, new_map_entry)
        else:
            dace_sdutils.canonicalize_memlet_trees_for_map(graph, containing_scope_or_none)
            dace_propagation.propagate_memlets_map_scope(sdfg, graph, containing_scope_or_none)

        # Workaround to ensure that some cache in DaCe has been cleared.
        # TODO(phimuell, iomaganaris): Before `hash_sdfg()` was used, but this was a
        #   very heavy operation. I think that this is enough.
        sdfg.reset_cfg_list()

        return new_map_entries_from_first, new_map_entries_from_second


@dace_properties.make_properties
class HorizontalSplitMapRange(SplitMapRange):
    """
    Identify overlapping range between parallel maps, and split the range in order
    to promote parallel map fusion. The transformation will merge the Maps that
    were created by the split among themselves.

    Note:
        The automatic fusion of Maps is only among the ones that were created, which
        is a bit different behaviour than for `VerticalSplitMapRange` where also
        possibilities between a newly created and an old map is considered.
    """

    first_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)
    second_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    fuse_map_fragments = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="If `True`, the transformation will try to fuse maps that have the same range together.",
    )
    never_consolidate_edges = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="If `True`, always create a new connector, instead of reusing one that referring to the same data.",
    )
    consolidate_edges_only_if_not_extending = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Only consolidate if this does not lead to an extension of the subset.",
    )

    _check_split_callback: Optional[HorizontalMapSplitCallback]

    def __init__(
        self,
        fuse_map_fragments: Optional[bool] = None,
        consolidate_edges_only_if_not_extending: Optional[bool] = None,
        never_consolidate_edges: Optional[bool] = None,
        check_split_callback: Optional[HorizontalMapSplitCallback] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if fuse_map_fragments is not None:
            self.fuse_map_fragments = fuse_map_fragments
        if consolidate_edges_only_if_not_extending is not None:
            self.consolidate_edges_only_if_not_extending = consolidate_edges_only_if_not_extending
        if never_consolidate_edges is not None:
            self.never_consolidate_edges = never_consolidate_edges
        self._check_split_callback = check_split_callback
        super().__init__(*args, **kwargs)

    @classmethod
    def expressions(cls) -> Any:
        map_fusion_parallel_match = dace_graph.OrderedMultiDiConnectorGraph()
        map_fusion_parallel_match.add_nodes_from([cls.first_map_entry, cls.second_map_entry])
        return [map_fusion_parallel_match]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        first_map_entry: dace_nodes.MapEntry = self.first_map_entry
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry
        first_map: dace_nodes.Map = first_map_entry.map
        second_map: dace_nodes.Map = second_map_entry.map

        if self._check_split_callback is not None:
            if not self._check_split_callback(self, first_map_entry, second_map_entry, graph, sdfg):
                return False

        # Ensure that the Maps are in the same scope.
        scope_dict = graph.scope_dict()
        if scope_dict[self.first_map_entry] is not scope_dict[self.second_map_entry]:
            return False

        # Test if the map is in the right scope.
        map_scope: Union[dace_nodes.Node, None] = scope_dict[self.first_map_entry]
        if self.only_toplevel_maps and (map_scope is not None):
            return False

        first_map_src_data = {
            iedge.src.label
            for iedge in graph.in_edges(self.first_map_entry)
            if isinstance(iedge.src, dace_nodes.AccessNode)
        }
        second_map_src_data = {
            iedge.src.label
            for iedge in graph.in_edges(self.second_map_entry)
            if isinstance(iedge.src, dace_nodes.AccessNode)
        }

        # Test if the Maps are parallel.
        if not dace_mfhelper.is_parallel(
            graph=graph, node1=first_map_entry, node2=second_map_entry
        ):
            return False

        if len(first_map_src_data.intersection(second_map_src_data)) == 0:
            # no common source access node
            return False

        splitted_range = gtx_mfutils.split_overlapping_map_range(first_map, second_map)
        if splitted_range is None:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        first_map_entry: dace_nodes.MapEntry = self.first_map_entry
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry

        # Now split the first and second Map.
        first_map_fragments, second_map_fragments = self.split_maps(
            graph,
            sdfg,
            first_map_entry,
            graph.exit_node(first_map_entry),
            second_map_entry,
            graph.exit_node(second_map_entry),
        )

        # If we do not want to fuse the maps, we can stop here.
        if not self.fuse_map_fragments:
            return

        # Look which Maps can be fused together. It is important that a Map can be listed multiple
        #  times. However, by fusing we remove them from the state, which is a problem. We solve
        #  this by always using "first Maps" as `to_node`, thus they survive. Furthermore, we
        #  only assign a second Map once.
        # TODO(phimuell, iomaganaris): Check and improve that.
        matched_map_fragments: list[tuple[dace_nodes.MapEntry, dace_nodes.MapEntry]] = []
        unmatched_second_map_fragments = second_map_fragments.copy()
        for first_map_fragment in first_map_fragments:
            for unmatched_second_map_fragment in unmatched_second_map_fragments.copy():
                if first_map_fragment.map.range == unmatched_second_map_fragment.map.range:
                    matched_map_fragments.append(
                        (first_map_fragment, unmatched_second_map_fragment)
                    )
                    unmatched_second_map_fragments.remove(unmatched_second_map_fragment)

        assert len(matched_map_fragments) > 0

        # We have to get the original scope dict before we start mutating the graph. Actually we
        #  would need to obtain the scope dict after every iteration again, which would require
        #  a rescan. But since the operations do not alter the scopes, at least not in a way that
        #  would affect us, we can be more efficient and get the thing at the beginning.
        scope_dict: Dict = graph.scope_dict()

        for first_map_fragment_entry, second_map_fragment_entry in matched_map_fragments:
            first_map_fragment_exit = graph.exit_node(first_map_fragment_entry)
            second_map_fragment_exit = graph.exit_node(second_map_fragment_entry)

            # Before we do anything we perform the renaming, i.e. we will rename the
            #  parameters of the second map such that they match the one of the first map.
            dace_mfhelper.rename_map_parameters(
                first_map=first_map_fragment_entry.map,
                second_map=second_map_fragment_entry.map,
                second_map_entry=second_map_fragment_entry,
                state=graph,
            )

            # Now we relocate all connectors from the second to the first map and remove
            #  the respective node of the second map.
            for to_node, from_node in [
                (first_map_fragment_entry, second_map_fragment_entry),
                (first_map_fragment_exit, second_map_fragment_exit),
            ]:
                dace_mfhelper.relocate_nodes(
                    from_node=from_node,
                    to_node=to_node,
                    state=graph,
                    sdfg=sdfg,
                    scope_dict=scope_dict,
                    never_consolidate_edges=self.never_consolidate_edges,
                    consolidate_edges_only_if_not_extending=self.consolidate_edges_only_if_not_extending,
                )
                # The relocate function does not remove the node, so we must do it.
                graph.remove_node(from_node)


@dace_properties.make_properties
class VerticalSplitMapRange(SplitMapRange):
    """
    Identify overlapping range between serial maps, and split the range in order
    to promote serial map fusion. Furthermore, this transformation also performs
    `SplitAccessNode` on all intermediates of the first and second Map.

    If `fuse_map_fragments` is set to `True`, the default is `False`, then the
    transformation will also perform MapFusionVertical, but limited to the newly
    created Maps.

    Args:
        only_toplevel_maps: Only applies to top level maps.
        check_split_callback: Callback used to check if a split should be performed
            or not, see `VerticalMapSplitCallback` for more.
        fuse_map_fragments: Immediately apply fusion and node splitting on the
            generated Maps.
        single_use_data: Use this as single use data and do not compute it on the fly.

    Note:
        - Even if `fuse_map_fragments` is `True` it might be that some fusion involving
            the split Maps is still possible, this is a bug.
        - Currently all AccessNodes between the first and second Map are subjected to
            `SplitAccessNode`, not just the ones that defined the split. This is a bug.
    """

    # Pattern Matching
    first_map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)
    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    second_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    fuse_map_fragments = dace_properties.Property(
        dtype=bool,
        default=False,  # NOTE: For backwards compatibility.
        desc="If `True`, the transformation will try to fuse the maps that were generated together.",
    )
    consolidate_edges_only_if_not_extending = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Only consolidate if this does not lead to an extension of the subset.",
    )

    _check_split_callback: Optional[VerticalMapSplitCallback]

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        check_split_callback: Optional[VerticalMapSplitCallback] = None,
        fuse_map_fragments: Optional[bool] = None,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        consolidate_edges_only_if_not_extending: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        if fuse_map_fragments is not None:
            self.fuse_map_fragments = fuse_map_fragments
        if consolidate_edges_only_if_not_extending is not None:
            self.consolidate_edges_only_if_not_extending = consolidate_edges_only_if_not_extending
        self._check_split_callback = check_split_callback
        self._single_use_data = single_use_data
        super().__init__(*args, **kwargs)

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.first_map_exit, cls.access_node, cls.second_map_entry
            ),
        ]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        """Check non overlapping range in the first and second map."""
        assert self.expr_index == expr_index

        first_map = self.first_map_exit.map
        first_map_entry: dace_nodes.MapEntry = graph.entry_node(self.first_map_exit)
        second_map = self.second_map_entry.map

        # Check the callback
        if self._check_split_callback is not None:
            if not self._check_split_callback(self, first_map, second_map, graph, sdfg):
                return False

        # Test if the map is in the right scope.
        map_scope: Union[dace_nodes.Node, None] = graph.scope_dict()[first_map_entry]
        if self.only_toplevel_maps and (map_scope is not None):
            return False

        if not self.access_node.desc(graph).transient:
            return False

        splitted_range = gtx_mfutils.split_overlapping_map_range(first_map, second_map)
        if splitted_range is None:
            return False

        # TODO(phimuell): Implement a check that the array is single use data?

        # Check that the data read by the second map from the intermediate AccessNode
        # are not copied to another AccessNode.
        access_node_out_edges = list(graph.out_edges(self.access_node))
        second_map_in_edges = list(graph.in_edges(self.second_map_entry))
        access_node_subset_into_second_map = [
            second_map_edge.data.src_subset
            for second_map_edge in second_map_in_edges
            if second_map_edge.src.data == self.access_node.data
            and second_map_edge.dst_conn.startswith("IN_")
        ]
        if any(
            any(
                gtx_dace_split.are_intersecting(out_edge.data.src_subset, matching_subset)
                for matching_subset in access_node_subset_into_second_map
            )
            for out_edge in access_node_out_edges
            if out_edge.dst != self.second_map_entry
            and isinstance(out_edge.dst, dace_nodes.AccessNode)
            and not out_edge.dst.desc(
                graph
            ).transient  # limit to transient nodes to avoid unnecessary splits in "vertically_implicit_solver_at_{corrector, predictor}_step"
        ):
            return False

        # Avoid spliting maps for fusion that if the second map has a nested map which is a neighbor reduction which
        # accesses the data that got generated by the first map. This is due to the fact that because of the neighbor
        # reduction, the second map might access a wider range of data that its range. This leads to a read-after-write
        # dependency between the first and second map.
        second_map_subgraph = graph.scope_subgraph(
            self.second_map_entry, include_entry=False, include_exit=False
        )
        for second_map_node in second_map_subgraph.nodes():
            if isinstance(second_map_node, dace_nodes.MapEntry):
                nested_map_input_edges_data = [
                    edge.data.data for edge in second_map_subgraph.in_edges(second_map_node)
                ]
                if self.access_node.data in nested_map_input_edges_data and any(
                    gtx_dace_args.is_connectivity_identifier(data)
                    for data in nested_map_input_edges_data
                ):
                    return False

        # TODO: Ensure that the fusion can be performed.

        return True

    def apply(self, graph: Union[dace.SDFGState, dace.SDFG], sdfg: dace.SDFG) -> None:
        """Split the map range in order to obtain an overlapping range between the first and second map."""

        first_map_entry: dace_nodes.MapEntry = graph.entry_node(self.first_map_exit)
        first_map_exit: dace_nodes.MapExit = self.first_map_exit
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry
        second_map_exit: dace_nodes.MapExit = graph.exit_node(self.second_map_entry)

        intermediates_that_might_need_splitting: list[dace_nodes.AccessNode] = []
        for oedge in graph.out_edges(first_map_exit):
            if isinstance(oedge.dst, dace_nodes.AccessNode) and any(
                ooedge.dst is second_map_entry for ooedge in graph.out_edges(oedge.dst)
            ):
                intermediates_that_might_need_splitting.append(oedge.dst)

        new_first_map_entries, new_second_map_entries = self.split_maps(
            graph,
            sdfg,
            first_map_entry,
            first_map_exit,
            second_map_entry,
            second_map_exit,
        )

        # We always split the AccessNode that are involved, regardless of `fuse_map_fragments`.
        # TODO(phimuell, iomaganaris): Refine this such that only the AccessNodes that
        #   defined the split are actually split and not some other random nodes.
        # TODO(phimuell): Make it possible to pass `self.single_use_data` to transformation.
        has_performed_a_split = False
        for intermediate_to_split in intermediates_that_might_need_splitting:
            if gtx_transformations.SplitAccessNode.can_be_applied_to(
                sdfg=sdfg,
                options={},
                access_node=intermediate_to_split,
            ):
                gtx_transformations.SplitAccessNode.apply_to(
                    sdfg=sdfg, options={}, access_node=intermediate_to_split
                )
                has_performed_a_split = True

        if not has_performed_a_split:
            # TODO(phimuell, iomaganaris): Is this an error?
            return

        # If requested perform the fusion.
        if self.fuse_map_fragments:
            assert len(intermediates_that_might_need_splitting) >= 1

            new_second_map_exits = [
                graph.exit_node(new_second_map_entry)
                for new_second_map_entry in new_second_map_entries
            ]

            def _restrict_fusion_to_newly_created_maps(
                this: gtx_transformations.MapFusionVertical,
                first_map_exit: dace_nodes.MapExit,
                second_map_entry: dace_nodes.MapEntry,
                state: dace.SDFGState,
                _sdfg: dace.SDFG,
            ) -> bool:
                # NOTE: For the check we have to use the entry of the first and the
                #   exit of the second Map, although there were not matched. The
                #   reason is that `MapFusionVertical` will remove the matched nodes,
                #   thus we will "loose sight of them". However, the "opposite" nodes
                #   of the Maps should survive, except in strange situation, but this
                #   is why we do not guarantee that everything related to the split
                #   has been merged.
                # NOTE: We use `or` here such that every fusion which involves
                #   a Map is is generated is considered. This should even work over
                #   chains, I guess.
                if (
                    state.entry_node(first_map_exit) in new_first_map_entries
                    or state.exit_node(second_map_entry) in new_second_map_exits
                ):
                    return True
                return False

            # NOTE: After here it is dangerous to access `self.PATTERN_NODE`.
            sdfg.reset_cfg_list()

            trafo = gtx_transformations.MapFusionVertical(
                check_fusion_callback=_restrict_fusion_to_newly_created_maps,
                only_toplevel_maps=self.only_toplevel_maps,
                consolidate_edges_only_if_not_extending=self.consolidate_edges_only_if_not_extending,
            )
            trafo._single_use_data = self._single_use_data

            # This is not efficient, but it is currently the only way to run it
            sdfg.apply_transformations_repeated(trafo, validate=False, validate_all=False)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.iterator.dispatcher import Dispatcher


builtin_dispatch = Dispatcher()


class BackendNotSelectedError(RuntimeError):
    def __init__(self) -> None:
        super().__init__("Backend not selected")


@builtin_dispatch
def as_fieldop(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def index(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def broadcast(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def deref(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def can_deref(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def shift(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def neighbors(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def map_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def make_const_list(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def list_get(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def lift(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def reduce(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def scan(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def cartesian_domain(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def unstructured_domain(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def named_range(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def if_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def cast_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def not_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def and_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def or_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def xor_(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def minus(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def plus(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def multiplies(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def divides(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def floordiv(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def eq(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def greater(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def less(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def less_equal(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def greater_equal(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def not_eq(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def make_tuple(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def tuple_get(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def abs(*args):  # noqa: A001 [builtin-variable-shadowing]
    raise BackendNotSelectedError()


@builtin_dispatch
def sin(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def cos(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def tan(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def arcsin(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def arccos(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def arctan(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def sinh(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def cosh(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def tanh(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def arcsinh(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def arccosh(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def arctanh(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def sqrt(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def exp(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def log(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def gamma(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def cbrt(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def floor(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def ceil(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def trunc(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def neg(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def isfinite(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def isinf(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def isnan(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def minimum(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def maximum(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def fmod(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def mod(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def power(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def int(*args):  # noqa: A001 [builtin-variable-shadowing]
    raise BackendNotSelectedError()


@builtin_dispatch
def int8(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def uint8(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def int16(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def uint16(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def int32(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def uint32(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def int64(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def uint64(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def float(*args):  # noqa: A001 [builtin-variable-shadowing]
    raise BackendNotSelectedError()


@builtin_dispatch
def float32(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def float64(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def bool(*args):  # noqa: A001 [builtin-variable-shadowing]
    raise BackendNotSelectedError()


@builtin_dispatch
def concat_where(*args):
    raise BackendNotSelectedError()


@builtin_dispatch
def get_domain_range(*args):
    raise BackendNotSelectedError()


UNARY_MATH_NUMBER_BUILTINS = {"abs", "neg"}
UNARY_LOGICAL_BUILTINS = {"not_"}
UNARY_MATH_FP_BUILTINS = {
    "sin",
    "cos",
    "tan",
    "arcsin",
    "arccos",
    "arctan",
    "sinh",
    "cosh",
    "tanh",
    "arcsinh",
    "arccosh",
    "arctanh",
    "sqrt",
    "exp",
    "log",
    "gamma",
    "cbrt",
    "floor",
    "ceil",
    "trunc",
}
UNARY_MATH_FP_PREDICATE_BUILTINS = {"isfinite", "isinf", "isnan"}
BINARY_MATH_NUMBER_BUILTINS = {
    "plus",
    "minus",
    "multiplies",
    "divides",
    "mod",
    "floordiv",  # TODO see https://github.com/GridTools/gt4py/issues/1136
    "minimum",
    "maximum",
    "fmod",
}
BINARY_MATH_COMPARISON_BUILTINS = {"eq", "less", "greater", "greater_equal", "less_equal", "not_eq"}
BINARY_LOGICAL_BUILTINS = {"and_", "or_", "xor_"}


#: builtin / dtype used to construct integer indices, like domain bounds
INTEGER_INDEX_BUILTIN = "int32"
INTEGER_TYPE_BUILTINS = {
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
}
FLOATING_POINT_TYPE_BUILTINS = {"float32", "float64"}
TYPE_BUILTINS = {*INTEGER_TYPE_BUILTINS, *FLOATING_POINT_TYPE_BUILTINS, "bool"}

ARITHMETIC_BUILTINS = {
    *UNARY_MATH_NUMBER_BUILTINS,
    *UNARY_LOGICAL_BUILTINS,
    *UNARY_MATH_FP_BUILTINS,
    *UNARY_MATH_FP_PREDICATE_BUILTINS,
    *BINARY_MATH_NUMBER_BUILTINS,
    "power",
    *BINARY_MATH_COMPARISON_BUILTINS,
    *BINARY_LOGICAL_BUILTINS,
}

BUILTINS = {
    "as_fieldop",  # `as_fieldop(stencil, domain)` creates field_operator from stencil (domain is optional, but for now required for embedded execution)
    "broadcast",
    "can_deref",
    "cartesian_domain",
    "cast_",
    "deref",
    "get_domain_range",
    "if_",
    "index",  # `index(dim)` creates a dim-field that has the current index at each point
    "shift",
    "list_get",
    "lift",
    "make_const_list",
    "make_tuple",
    "map_",
    "named_range",
    "neighbors",
    "reduce",
    "scan",
    "tuple_get",
    "unstructured_domain",
    "concat_where",
    *ARITHMETIC_BUILTINS,
    *TYPE_BUILTINS,
}

__all__ = [*BUILTINS]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from .npir_codegen import NpirCodegen
from .oir_to_npir import OirToNpir
from .scalars_to_temps import ScalarsToTemporaries


__all__ = ["NpirCodegen", "OirToNpir", "ScalarsToTemporaries"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
from typing import Any, Optional, Union

import dace
from dace import properties as dace_properties, transformation as dace_transformation
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation import helpers as dace_helpers

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dace_properties.make_properties
class FuseHorizontalConditionBlocks(dace_transformation.SingleStateTransformation):
    """Fuses two conditional blocks that share the same condition variable and are
    not dependent to each other (i.e. the output of one of them is used as input to the other)
    into a single conditional block.
    The motivation for this transformation is to reduce the number of conditional blocks
    which generate if statements in the CPU or GPU code, which can lead to performance improvements.
    Example:
    Before fusion:
    ```
    if __cond:
        __output1 = __arg1 * 2.0
    else:
        __output1 = __arg2 + 3.0
    if __cond:
        __output2 = __arg3 - 1.0
    else:
        __output2 = __arg4 / 4.0
    ```
    After fusion:
    ```
    if __cond:
        __output1 = __arg1 * 2.0
        __output2 = __arg3 - 1.0
    else:
        __output1 = __arg2 + 3.0
        __output2 = __arg4 / 4.0
    ```
    """

    conditional_access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    nsdfg_a = dace_transformation.PatternNode(dace_nodes.NestedSDFG)
    nsdfg_b = dace_transformation.PatternNode(dace_nodes.NestedSDFG)

    # The fusion of the two conditional blocks can happen in any order. To avoid any indeterminism distinguish which one is the fused and which one is the extended conditional block which will include the fused one.
    @staticmethod
    def _order_conditional_blocks_based_on_label(
        conditional_block_0: dace.sdfg.state.ConditionalBlock,
        conditional_block_1: dace.sdfg.state.ConditionalBlock,
    ) -> tuple[dace.sdfg.state.ConditionalBlock, dace.sdfg.state.ConditionalBlock]:
        if conditional_block_0.label < conditional_block_1.label:
            return conditional_block_0, conditional_block_1
        else:
            return conditional_block_1, conditional_block_0

    @staticmethod
    def _order_conditional_blocks_based_on_number_of_common_branches_and_label(
        conditional_block_0: dace.sdfg.state.ConditionalBlock,
        conditional_block_1: dace.sdfg.state.ConditionalBlock,
    ) -> tuple[dace.sdfg.state.ConditionalBlock, dace.sdfg.state.ConditionalBlock] | None:
        branches_0 = conditional_block_0.branches
        branches_1 = conditional_block_1.branches

        # We compute a set of the conditions because we don't care about their order since all the
        # conditions should either be based in a scalar boolean (`__cond`) or be None which means
        # the `else` condition.
        branch_0_conditions = {branch[0].as_string if branch[0] else None for branch in branches_0}
        branch_1_conditions = {branch[0].as_string if branch[0] else None for branch in branches_1}

        intersection_of_conditions = branch_0_conditions.intersection(branch_1_conditions)

        # No matching conditions found between the two `ConditionalBlock`s
        if not intersection_of_conditions:
            return None
        # Matching conditions between the two `ConditionalBlock`s is the same as the number of branches in both `ConditionalBlock`s.
        # This means that all branches have matching conditions and thus we can use the labels to distinguish which one is the fused and which one is the extended conditional block.
        elif (
            len(intersection_of_conditions) == len(branch_0_conditions) == len(branch_1_conditions)
        ):
            return FuseHorizontalConditionBlocks._order_conditional_blocks_based_on_label(
                conditional_block_0, conditional_block_1
            )
        # This means that all the branches of `conditional_block_1` match some or all of the conditions of `conditional_block_0` so we can merge `conditional_block_1` into `conditional_block_0`.
        elif len(intersection_of_conditions) == len(branch_1_conditions):
            return conditional_block_0, conditional_block_1
        # This means that all the branches of `conditional_block_0` match some or all of the conditions of `conditional_block_1` so we can merge `conditional_block_0`` into `conditional_block_1`.
        elif len(intersection_of_conditions) == len(branch_0_conditions):
            return conditional_block_1, conditional_block_0
        return None

    @classmethod
    def expressions(cls) -> Any:
        conditionalblock_fusion_parallel_match = dace_graph.OrderedMultiDiConnectorGraph()
        conditionalblock_fusion_parallel_match.add_nedge(
            cls.conditional_access_node, cls.nsdfg_a, dace.Memlet()
        )
        conditionalblock_fusion_parallel_match.add_nedge(
            cls.conditional_access_node, cls.nsdfg_b, dace.Memlet()
        )
        return [conditionalblock_fusion_parallel_match]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        conditional_access_node: dace_nodes.AccessNode = self.conditional_access_node
        conditional_access_node_desc = conditional_access_node.desc(sdfg)
        nsdfg_a = self.nsdfg_a
        nsdfg_b = self.nsdfg_b

        scope_dict = graph.scope_dict()

        # Check that the common access node is a boolean scalar
        if (
            not isinstance(conditional_access_node_desc, dace.data.Scalar)
            or conditional_access_node_desc.dtype != dace.bool_
        ):
            return False

        # Check that both conditional blocks are in the same parent SDFG
        if nsdfg_a.sdfg.parent is not nsdfg_b.sdfg.parent:
            return False

        # Check that the nested SDFGs' names starts with "if_stmt"
        if not (
            nsdfg_a.sdfg.name.startswith("if_stmt") and nsdfg_b.sdfg.name.startswith("if_stmt")
        ):
            return False

        # Make sure that the nested SDFGs contain only one conditional block each
        if nsdfg_a.sdfg.number_of_nodes() != 1 or nsdfg_b.sdfg.number_of_nodes() != 1:
            return False

        if not (
            isinstance(nsdfg_a.sdfg.nodes()[0], dace.sdfg.state.ConditionalBlock)
            and isinstance(nsdfg_b.sdfg.nodes()[0], dace.sdfg.state.ConditionalBlock)
        ):
            return False

        conditional_block_a = nsdfg_a.sdfg.nodes()[0]
        conditional_block_b = nsdfg_b.sdfg.nodes()[0]

        # This transformation accepts only a boolean condition. This means that the conditions of
        # the branches can only be 2 or 1. Either `__cond` and `(not __cond)` or `__cond` and `None`
        # or just `__cond` or `(not __cond)`.
        if len(conditional_block_a.branches) > 2 or len(conditional_block_b.branches) > 2:
            return False

        # Make sure that the branches of both `ConditionalBlock`s have only one state
        # TODO(iomaganaris): In case there are more states we can find a way to add one after the other
        for conditional_block in [conditional_block_a, conditional_block_b]:
            for control_flow_region in conditional_block.sub_regions():
                if len(control_flow_region.nodes()) != 1 or not isinstance(
                    control_flow_region.nodes()[0], dace.sdfg.state.SDFGState
                ):
                    return False

        conditional_block_tuple = (
            self._order_conditional_blocks_based_on_number_of_common_branches_and_label(
                conditional_block_a, conditional_block_b
            )
        )
        if not conditional_block_tuple:
            return False
        extended_conditional_block, fused_conditional_block = conditional_block_tuple

        nested_sdfg_of_extended_conditional_block = (
            extended_conditional_block.sdfg.parent_nsdfg_node
        )
        nested_sdfg_of_fused_conditional_block = fused_conditional_block.sdfg.parent_nsdfg_node

        # Check that the symbol mappings are compatible. If there's a symbol that is in both mappings but mapped to different definitions then we skip fusing the conditional blocks.
        # TODO(iomaganaris): One could also rename the symbols instead of skipping the fusion but for now we keep it simple
        sym_map1 = nested_sdfg_of_extended_conditional_block.symbol_mapping
        sym_map2 = nested_sdfg_of_fused_conditional_block.symbol_mapping
        if any(str(sym_map1[sym]) != str(sym_map2[sym]) for sym in sym_map2 if sym in sym_map1):
            return False

        # Make sure that both conditional blocks are in the same scope
        if (
            scope_dict[nested_sdfg_of_extended_conditional_block]
            != scope_dict[nested_sdfg_of_fused_conditional_block]
        ):
            return False

        # Make sure that both conditional blocks are in a map scope
        if not isinstance(
            scope_dict[nested_sdfg_of_extended_conditional_block], dace.nodes.MapEntry
        ):
            return False

        # Check that there is an edge to the conditional blocks with dst_conn == "__cond"
        cond_edges_first = [
            e
            for e in graph.in_edges(nested_sdfg_of_extended_conditional_block)
            if e.dst_conn and e.dst_conn == "__cond"
        ]
        if len(cond_edges_first) != 1:
            return False
        cond_edges_second = [
            e
            for e in graph.in_edges(nested_sdfg_of_fused_conditional_block)
            if e.dst_conn and e.dst_conn == "__cond"
        ]
        if len(cond_edges_second) != 1:
            return False
        cond_edge_first = cond_edges_first[0]
        cond_edge_second = cond_edges_second[0]
        if cond_edge_first.data.is_empty() or cond_edge_second.data.is_empty():
            return False
        if not all(
            cond_edge.src is conditional_access_node
            for cond_edge in [cond_edge_first, cond_edge_second]
        ):
            return False

        # Need to check also that first and second nested SDFGs are not reachable from each other
        if gtx_transformations.utils.is_reachable(
            start=nested_sdfg_of_extended_conditional_block,
            target=nested_sdfg_of_fused_conditional_block,
            state=graph,
        ) or gtx_transformations.utils.is_reachable(
            start=nested_sdfg_of_fused_conditional_block,
            target=nested_sdfg_of_extended_conditional_block,
            state=graph,
        ):
            return False

        return True

    def apply(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> None:
        conditional_access_node: dace_nodes.AccessNode = self.conditional_access_node
        nsdfg_a = self.nsdfg_a
        nsdfg_b = self.nsdfg_b
        assert nsdfg_a.sdfg.number_of_nodes() == 1 and nsdfg_b.sdfg.number_of_nodes() == 1
        conditional_block_a = nsdfg_a.sdfg.nodes()[0]
        conditional_block_b = nsdfg_b.sdfg.nodes()[0]
        conditional_block_tuple = (
            self._order_conditional_blocks_based_on_number_of_common_branches_and_label(
                conditional_block_a, conditional_block_b
            )
        )
        assert conditional_block_tuple is not None
        extended_conditional_block, fused_conditional_block = conditional_block_tuple

        nested_sdfg_of_extended_conditional_block = (
            extended_conditional_block.sdfg.parent_nsdfg_node
        )
        nested_sdfg_of_fused_conditional_block = fused_conditional_block.sdfg.parent_nsdfg_node

        # Copy missing symbols from second conditional block to the first one.
        #  For the symbols that are already in `nested_sdfg_of_extended_conditional_block.symbol_mapping` we know
        #  that the definition matches. Thus there is no need to perform symbol
        #  renaming.
        missing_symbols = {
            sym2: val2
            for sym2, val2 in nested_sdfg_of_fused_conditional_block.symbol_mapping.items()
            if sym2 not in nested_sdfg_of_extended_conditional_block.symbol_mapping
        }
        for missing_symb, symb_def in missing_symbols.items():
            nested_sdfg_of_extended_conditional_block.symbol_mapping[missing_symb] = symb_def
            nested_sdfg_of_extended_conditional_block.sdfg.add_symbol(
                missing_symb,
                nested_sdfg_of_fused_conditional_block.sdfg.symbols[missing_symb],
                find_new_name=False,
            )

        # Store number of original arrays to check later that all the necessary arrays have been moved
        total_original_arrays = len(extended_conditional_block.sdfg.arrays) + len(
            fused_conditional_block.sdfg.arrays
        )

        # Store the new names for the arrays of the second conditional block (transients and globals) to avoid name clashes and add their data descriptors
        # to the first conditional block SDFG. We don't have to add `__cond` because we know it's the same for both conditional blocks.
        # TODO(iomaganaris): Remove inputs to the conditional block that come from the same AccessNodes (same data)
        second_arrays_rename_map: dict[str, str] = {}
        for data_name, data_desc in fused_conditional_block.sdfg.arrays.items():
            if data_name == "__cond":
                continue
            new_data_name = gtx_transformations.utils.unique_name(data_name) + "_from_cb_fusion"
            data_desc_renamed = copy.deepcopy(data_desc)
            second_arrays_rename_map[data_name] = (
                nested_sdfg_of_extended_conditional_block.sdfg.add_datadesc(
                    new_data_name, data_desc_renamed, find_new_name=True
                )
            )

        # Move the connectors from the second conditional block to the first
        # TODO(iomaganaris): Here we copy empty memlets used for scheduling as well. This means that the first conditional blocks inherits the scheduling of the second one as well. Maybe that's not good in some cases to hide latency but for now we keep it as it is
        for edge in graph.in_edges(nested_sdfg_of_fused_conditional_block):
            if edge.dst_conn == "__cond":
                continue
            nested_sdfg_of_extended_conditional_block.add_in_connector(
                second_arrays_rename_map[edge.dst_conn]
            )
            dace_helpers.redirect_edge(
                state=graph,
                edge=edge,
                new_dst_conn=second_arrays_rename_map[edge.dst_conn],
                new_dst=nested_sdfg_of_extended_conditional_block,
            )
        for edge in graph.out_edges(nested_sdfg_of_fused_conditional_block):
            nested_sdfg_of_extended_conditional_block.add_out_connector(
                second_arrays_rename_map[edge.src_conn]
            )
            dace_helpers.redirect_edge(
                state=graph,
                edge=edge,
                new_src_conn=second_arrays_rename_map[edge.src_conn],
                new_src=nested_sdfg_of_extended_conditional_block,
            )

        def _find_corresponding_branch_in_fused(
            extended_branch: tuple[
                Optional[dace.sdfg.state.CodeBlock], dace.sdfg.state.ControlFlowRegion
            ],
            fused_conditional_block: dace.sdfg.state.ConditionalBlock,
        ) -> dace.sdfg.state.ControlFlowRegion | None:
            extended_branch_condition = extended_branch[0].as_string if extended_branch[0] else None
            for branch in fused_conditional_block.branches:
                branch_condition = branch[0].as_string if branch[0] else None
                if branch_condition == extended_branch_condition:
                    return branch
            return None

        # Copy first the nodes from the second conditional block to the first
        # Create a dictionary that maps the original nodes in the second conditional
        # block to the new nodes in the first conditional block to be able to properly connect the edges later
        nodes_renamed_map: dict[dace_nodes.Node, dace_nodes.Node] = {}
        for branch in extended_conditional_block.branches:
            corresponding_branch_in_second = _find_corresponding_branch_in_fused(
                branch, fused_conditional_block
            )
            if not corresponding_branch_in_second:
                continue
            corresponding_state_in_second = corresponding_branch_in_second[1].nodes()[0]
            first_inner_state = branch[1].nodes()[0]
            # Save edges of second conditional block to a state to be able to delete the nodes from the second conditional block
            edges_to_copy = list(corresponding_state_in_second.edges())
            nodes_to_move = list(corresponding_state_in_second.nodes())
            for node in nodes_to_move:
                new_node = node
                if isinstance(node, dace_nodes.AccessNode):
                    new_data_name = second_arrays_rename_map[node.data]
                    new_node = dace_nodes.AccessNode(new_data_name)
                nodes_renamed_map[node] = new_node
                # Remove the original node from the second conditional block to avoid any potential issues
                # with the nodes coexisting in two states
                corresponding_state_in_second.remove_node(node)
                first_inner_state.add_node(new_node)

            for edge_to_copy in edges_to_copy:
                new_edge = first_inner_state.add_edge(
                    nodes_renamed_map[edge_to_copy.src],
                    edge_to_copy.src_conn,
                    nodes_renamed_map[edge_to_copy.dst],
                    edge_to_copy.dst_conn,
                    edge_to_copy.data,
                )
                if not new_edge.data.is_empty():
                    new_edge.data.data = second_arrays_rename_map[new_edge.data.data]

        for edge in list(graph.out_edges(conditional_access_node)):
            if edge.dst == nested_sdfg_of_fused_conditional_block:
                graph.remove_edge(edge)

        # TODO(iomaganaris): Atm need to remove both references to remove NestedSDFG from graph
        #  fused_conditional_block is inside the SDFG of NestedSDFG nested_sdfg_of_fused_conditional_block and removing only
        #  one of them keeps a reference to the other one so none is properly deleted from the SDFG.
        #  For now remove both but maybe this can be improved in the future.
        graph.remove_node(fused_conditional_block)
        graph.remove_node(nested_sdfg_of_fused_conditional_block)

        new_arrays = len(nested_sdfg_of_extended_conditional_block.sdfg.arrays)
        assert new_arrays == total_original_arrays - 1, (
            f"After fusion, expected {total_original_arrays - 1} arrays but found {new_arrays}"
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import ir


class NormalizeShifts(PreserveLocationVisitor, NodeTranslator):
    def visit_FunCall(self, node: ir.FunCall):
        node = self.generic_visit(node)
        if (
            isinstance(node.fun, ir.FunCall)
            and isinstance(node.fun.fun, ir.SymRef)
            and node.fun.fun.id == "shift"
            and node.args
            and isinstance(node.args[0], ir.FunCall)
            and isinstance(node.args[0].fun, ir.FunCall)
            and isinstance(node.args[0].fun.fun, ir.SymRef)
            and node.args[0].fun.fun.id == "shift"
        ):
            # shift(args1...)(shift(args2...)(it)) -> shift(args2..., args1...)(it)
            assert len(node.args) == 1
            return ir.FunCall(
                fun=ir.FunCall(
                    fun=ir.SymRef(id="shift"), args=node.args[0].fun.args + node.fun.args
                ),
                args=node.args[0].args,
            )
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Any, Optional

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    subsets as dace_sbs,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes, utils as dace_sdutils
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations
from gt4py.next.program_processors.runners.dace.transformations import (
    map_fusion_utils as gtx_mfutils,
    splitting_tools as gtx_stools,
)


@dace_properties.make_properties
class MapSplitter(dace_transformation.SingleStateTransformation):
    """Split a Map based on a transient.

    The transformation matches the following pattern `MapExit -> (AccessNode)`,
    where `AccessNode` needs to be a transient single use data whose only producer
    is the matched Map and has only a single consumer. Furthermore, more data
    is written into `AccessNode` than is read from it. In that case the
    producing Map and `AccessNode` is split into fragments, such that a fragment
    either generates data that is fully consumed or not at all.
    Fragments that generate data that is not used are removed, unless
    `remove_dead_dataflow` is set to `False`.

    Args:
        removed_dead_dataflow: If `True`, the default, remove dataflow that is not
            read anywhere.
        single_use_data: Use this to classify single use data, see `dace.FindSingleUseData()`.

    Note:
        - The "more data is written into `AccessNode` than read from it" technically
            violates ADR18, but is the result of some broadcast expressions in
            relation to `concat_where`.
    """

    map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)
    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    remove_dead_dataflow = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="Remove dead dataflow directly.",
    )

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.map_exit, cls.access_node)]

    def __init__(
        self,
        remove_dead_dataflow: Optional[bool] = None,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._single_use_data = single_use_data
        if remove_dead_dataflow is not None:
            self.remove_dead_dataflow = remove_dead_dataflow
        super().__init__(*args, **kwargs)

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        # We assume that the Map only generates a single data.
        map_exit: dace_nodes.MapExit = self.map_exit
        if graph.out_degree(map_exit) != 1:
            return False

        # The AccessNode must be single use data and have only one producer. Single use
        #  data is tested at the very end, to avoid computing it too often.
        # NOTE: Requiring that `self.access_node` is a transient is mainly to simplify
        #   the implementation, but the transient should always exists anyway.
        access_node: dace_nodes.AccessNode = self.access_node
        ac_desc: dace_data.Data = access_node.desc(sdfg)
        if not ac_desc.transient:
            return False
        if graph.in_degree(access_node) != 1:
            return False

        # For simplicity we assume that the AccessNode only has a single consumer.
        #  This makes handling a bit simpler, as we do not have to compose the
        #  subset that is read.
        # TODO(phimuell): Remove this requirement.
        if graph.out_degree(access_node) != 1:
            return False

        map_ac_edge = next(iter(graph.in_edges(access_node)))
        ac_consumer_edge = next(iter(graph.out_edges(access_node)))
        produced_subset: dace_sbs.Range = map_ac_edge.data.get_dst_subset(map_ac_edge, graph)
        consumed_subset: dace_sbs.Range = ac_consumer_edge.data.get_src_subset(
            ac_consumer_edge, graph
        )

        # This simplifies the implementation of the correction.
        # TODO(phimuell): Not sure how useful lifting this restriction would be.
        if len(map_exit.map.params) != produced_subset.dims():
            return False
        if not all(
            (sbs[0] == 0) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
            for sbs in produced_subset
        ):
            return False
        if not all(
            (psize == msize) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
            for psize, msize in zip(produced_subset.size(), map_exit.map.range.size())
        ):
            return False

        # In order for this transformation to make sense, we require that more data is
        #  written to the AccessNode than read from it.
        if consumed_subset == produced_subset:
            return False
        if not produced_subset.covers(consumed_subset):
            return False

        # We now test if the AccessNode is a single used data. We do it here such that
        #  we can postpone the scanning of the SDFG as long as possible.
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        if access_node.data not in single_use_data[sdfg]:
            return False

        return True

    def apply(self, graph: dace.SDFGState, sdfg: dace.SDFG) -> None:
        map_exit: dace_nodes.MapExit = self.map_exit
        map_entry: dace_nodes.MapEntry = graph.entry_node(map_exit)
        access_node: dace_nodes.AccessNode = self.access_node

        map_ac_edge = next(iter(graph.in_edges(access_node)))
        ac_consumer_edge = next(iter(graph.out_edges(access_node)))
        assert graph.in_degree(access_node) == 1 and graph.out_degree(access_node) == 1

        produced_subset: dace_sbs.Range = map_ac_edge.data.get_dst_subset(map_ac_edge, graph)
        consumed_subset: dace_sbs.Range = ac_consumer_edge.data.get_src_subset(
            ac_consumer_edge, graph
        )

        # Now decompose the production subset, we will use this information to "split"
        #  the Map.
        # NOTE: Name swapping is intentional.
        split_production_subsets = gtx_stools.decompose_subset(
            consumer=produced_subset, producer=consumed_subset
        )
        assert split_production_subsets is not None and len(split_production_subsets) > 1

        # Now create new Maps of the appropriate ranges.
        # NOTE: `split_production_subsets` is only based on what is accessed at
        #   `self.access_node`, however, we want to split the Map based on it.
        #   The main problem is if there are offsets, i.e. `output[__i - 3]`.
        #   To solve this we realize that while the actual decomposed ranges are
        #   meaningless, they still have the same size. Thus we simply correct them.
        map_start_index = [map_exit.map.range[i][0] for i in range(len(map_exit.map.range))]
        new_map_exits: list[dace_nodes.MapExit] = []
        for i, production_subset in enumerate(split_production_subsets):
            sub_me, sub_mx = gtx_mfutils.copy_map_graph_with_new_range(
                sdfg=sdfg,
                state=graph,
                map_entry=map_entry,
                map_exit=map_exit,
                map_range=production_subset.offset_new(map_start_index, negative=False),
                suffix=str(i),
            )
            # `copy_map_graph_with_new_range()` just copies all Memlets, thus the
            #  subsets that are read and written by the new Maps are still the same
            #  as the original one. We must fix that otherwise `SplitAccessNode` will
            #  fail.
            # TODO(phimuell): Find out how to limit propagation only to the Maps that
            #   are not dead dataflow, in case we remove them anyway.
            dace_sdutils.canonicalize_memlet_trees_for_map(state=graph, map_node=sub_me)
            dace.sdfg.propagation.propagate_memlets_map_scope(
                sdfg=sdfg,
                state=graph,
                map_entry=sub_me,
            )
            new_map_exits.append(sub_mx)
        gtx_mfutils.delete_map(graph=graph, map_entry=map_entry, map_exit=map_exit)

        # Now perform the split of `self.access_node`. Parts that are not read will
        #  later be eliminated by dead dataflow elimination.
        gtx_transformations.SplitAccessNode.apply_to(
            sdfg=sdfg,
            options={
                "assume_single_use_data": True,  # Was tested in `can_be_applied()`.
            },
            verify=True,
            access_node=access_node,
        )

        if not self.remove_dead_dataflow:
            return

        # Because `self.access_node` is single use data, we know that if the split
        #  node is not directly used, i.e. has an output degree of 0, that it is dead
        #  dataflow and we can remove it.
        for mx in new_map_exits:
            output_node = next(iter(graph.out_edges(mx))).dst
            assert graph.out_degree(mx) == 1
            assert (
                isinstance(output_node, dace_nodes.AccessNode) and output_node.desc(sdfg).transient
            )

            if graph.out_degree(output_node) == 0:
                gtx_transformations.gt_remove_map(
                    sdfg=sdfg,
                    state=graph,
                    map_entry=graph.entry_node(mx),
                    # We would need to update single use data to really use it.
                    remove_unused_data=False,
                )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implements the lowering of scan field operator.

This builtin translator implements the `PrimitiveTranslator` protocol as other
translators in `gtir_to_sdfg_primitives` module. This module implements the scan
translator, separately from the `gtir_to_sdfg_primitives` module, because the
parsing of input arguments as well as the construction of the map scope differ
from a regular field operator, which requires slightly different helper methods.
Besides, the function code is quite large, another reason to keep it separate
from other translators.

The current GTIR representation of the scan operator is based on iterator view.
This is likely to change in the future, to enable GTIR optimizations for scan.
"""

from __future__ import annotations

from typing import Iterable, Sequence

import dace
from dace import nodes as dace_nodes, subsets as dace_subsets

from gt4py import eve
from gt4py.eve.extended_typing import MaybeNestedInTuple
from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
)
from gt4py.next.iterator.transforms import infer_domain
from gt4py.next.program_processors.runners.dace.lowering import (
    gtir_dataflow,
    gtir_domain,
    gtir_to_sdfg,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
)
from gt4py.next.type_system import type_info as ti, type_specifications as ts


def _parse_scan_fieldop_arg(
    node: gtir.Expr,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    field_domain: gtir_domain.FieldopDomain,
) -> MaybeNestedInTuple[gtir_dataflow.MemletExpr]:
    """Helper method to visit an expression passed as argument to a scan field operator.

    On the innermost level, a scan operator is lowered to a loop region which computes
    column elements in the vertical dimension.

    It differs from the helper method `gtir_to_sdfg_primitives` in that field arguments
    are passed in full shape along the vertical dimension, rather than as iterator.
    """

    def _parse_fieldop_arg_impl(
        arg: gtir_to_sdfg_types.FieldopData,
    ) -> gtir_dataflow.MemletExpr:
        arg_expr = arg.get_local_view(field_domain, ctx.sdfg)
        if isinstance(arg_expr, gtir_dataflow.MemletExpr):
            return arg_expr
        # In scan field operator, the arguments to the vertical stencil are passed by value.
        # Therefore, the full field shape is passed as `MemletExpr` rather than `IteratorExpr`.
        field_type = ts.FieldType(
            dims=[dim for dim, _ in arg_expr.field_domain], dtype=arg_expr.gt_dtype
        )
        return gtir_dataflow.MemletExpr(
            arg_expr.field, field_type, arg_expr.get_memlet_subset(ctx.sdfg)
        )

    arg = sdfg_builder.visit(node, ctx=ctx)

    if isinstance(arg, gtir_to_sdfg_types.FieldopData):
        return _parse_fieldop_arg_impl(arg)
    else:
        # handle tuples of fields
        return gtx_utils.tree_map(_parse_fieldop_arg_impl)(arg)


def _create_scan_field_operator_impl(
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    output_edge: gtir_dataflow.DataflowOutputEdge | None,
    output_domain: infer_domain.NonTupleDomainAccess,
    output_type: ts.FieldType,
    map_exit: dace_nodes.MapExit | None,
) -> gtir_to_sdfg_types.FieldopData | None:
    """
    Helper method to allocate a temporary array that stores one field computed
    by the scan field operator.

    This method is called by `_create_scan_field_operator()`.

    Similar to `gtir_to_sdfg_primitives._create_field_operator_impl()` but
    for scan field operators. It differs in that the scan loop region produces
    a field along the vertical dimension, rather than a single point.
    Therefore, the memlet subset will write a slice into the result array, that
    corresponds to the full vertical shape for each horizontal grid point.

    Another difference is that this function is called on all fields inside a tuple,
    in case of tuple return. Note that a regular field operator only computes a
    single field, never a tuple of fields. For tuples, it can happen that one of
    the nested fields is not used, outside the scan field operator, and therefore
    does not need to be computed. Then, the domain inferred by gt4py on this field
    is empty and the corresponding `output_edge` argument to this function is None.
    In this case, the function does not allocate an array node for the output field
    and returns None.

    Refer to `gtir_to_sdfg_primitives._create_field_operator_impl()` for
    the description of function arguments.
    """
    if output_edge is None:
        # According to domain inference, this tuple field does not need to be computed.
        assert output_domain == infer_domain.DomainAccessDescriptor.NEVER
        return None
    assert isinstance(output_domain, domain_utils.SymbolicDomain)
    field_domain = gtir_domain.get_field_domain(output_domain)

    outer_output_desc = output_edge.result.dc_node.desc(ctx.sdfg)
    assert isinstance(outer_output_desc, dace.data.Array)

    if isinstance(output_edge.result.gt_dtype, ts.ScalarType):
        assert isinstance(output_type.dtype, ts.ScalarType)
        if output_edge.result.gt_dtype != output_type.dtype:
            raise TypeError(
                f"Type mismatch, expected {output_type.dtype} got {output_edge.result.gt_dtype}."
            )
        # the scan field operator computes a column of scalar values
        assert len(outer_output_desc.shape) == 1
    else:
        raise NotImplementedError("scan with list output is not supported")

    # the memory layout of the output field follows the field operator compute domain
    field_dims, field_origin, field_shape = gtir_domain.get_field_layout(field_domain)
    field_subset = gtir_domain.get_element_subset(field_dims, field_origin)

    # the vertical dimension used as scan column is computed by the `LoopRegion`
    # inside the map scope, therefore it is excluded from the map range
    scan_dim_index = [sdfg_builder.is_column_axis(dim) for dim in field_dims].index(True)

    # The map scope writes the full-shape dimension corresponding to the scan column.
    field_subset = (
        dace_subsets.Range(field_subset[:scan_dim_index])
        + dace_subsets.Range.from_string(f"0:{outer_output_desc.shape[0]}")
        + dace_subsets.Range(field_subset[scan_dim_index + 1 :])
    )

    # Create the final data storage, that is outside of the surrounding Map.
    field_name, field_desc = sdfg_builder.add_temp_array(
        ctx.sdfg, field_shape, outer_output_desc.dtype
    )
    field_node = ctx.state.add_access(field_name)

    # Now connect the output connector on the nested SDFG with the result field
    #  outside the scan map scope. For 1D domain, containing only the column dimension,
    #  there is no map scope, since the map range is only on the horizontal domain.
    #  Up to now the nested SDFG is writing into a transient data container that
    #  has the size to hold one column. The function below, that does the connection,
    #  will remove that transient and write directly to the result field.
    inner_map_output_temporary_removed = output_edge.connect(map_exit, field_node, field_subset)
    if not inner_map_output_temporary_removed:
        raise ValueError("The scan nested SDFG is expected to write directly to the result field.")

    assert ctx.state.in_degree(field_node) == 1
    field_node_path = ctx.state.memlet_path(next(iter(ctx.state.in_edges(field_node))))
    assert field_node_path[-1].dst is field_node

    # The temporary node which the nested SDFG was writing to has been deleted,
    #  and the nested SDFG will write directly to the result field. Thus, we have
    #  to modify the stride of the scan column array inside the nested SDFG to match
    #  the stride outside in the corresponding field dimension.
    nsdfg_scan = field_node_path[0].src
    assert isinstance(nsdfg_scan, dace_nodes.NestedSDFG)
    inner_output_name = field_node_path[0].src_conn
    inner_output_desc = nsdfg_scan.sdfg.arrays[inner_output_name]
    assert len(inner_output_desc.shape) == 1

    if isinstance(inner_output_desc, dace.data.Array):
        # The result field on the outside is a transient array, allocated inside this
        # function, so we know that its stride is constant. We just need to set it on
        # the inside array, and we do not need to map any stride symbol.
        outside_output_stride = field_desc.strides[scan_dim_index]
        assert str(outside_output_stride).isdigit()
        inner_output_desc.set_shape(inner_output_desc.shape, strides=(outside_output_stride,))
    else:
        # Special case where we only write the last level of the scan column.
        assert isinstance(inner_output_desc, dace.data.Scalar)
        assert bool(field_shape[scan_dim_index] == 1)

    return gtir_to_sdfg_types.FieldopData(
        field_node, ts.FieldType(field_dims, output_edge.result.gt_dtype), tuple(field_origin)
    )


def _create_scan_field_operator(
    ctx: gtir_to_sdfg.SubgraphContext,
    field_domain: gtir_domain.FieldopDomain,
    node_type: ts.FieldType | ts.TupleType,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    input_edges: Iterable[gtir_dataflow.DataflowInputEdge],
    output: MaybeNestedInTuple[gtir_dataflow.DataflowOutputEdge | None],
    output_domain: infer_domain.DomainAccess,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Helper method to build the output of a field operator, which can consist of
    a single field or a tuple of fields.

    Similar to `gtir_to_sdfg_primitives._create_field_operator()` but for scan
    field operators. The main difference is that the scan vertical dimension is
    excluded from the map range. This because the vertical dimension is traversed
    by a loop region in a mapped nested SDFG.

    Refer to `gtir_to_sdfg_primitives._create_field_operator()` for the
    description of function arguments. Note that the return value is different,
    because the scan field operator can return a tuple of fields, while a regular
    field operator return a single field. The domain of the nested fields, in
    a tuple, can be empty, in case the nested field is not used outside the scan.
    In this case, the corresponding `output` edge will be None and this function
    will also return None for the corresponding field inside the tree-like result.
    """
    dims, _, _ = gtir_domain.get_field_layout(field_domain)

    # create a map scope to execute the `LoopRegion` over the horizontal domain
    if len(dims) == 1:
        # We construct the scan field operator on the horizontal domain, while the
        # vertical dimension (the column axis) is computed by the loop region.
        # If the field operator computes only the column axis (a 1d scan field operator),
        # there is no horizontal domain, therefore the map scope is not needed.
        # This case currently produces wrong CUDA code because of a DaCe issue
        # (see https://github.com/GridTools/gt4py/issues/1136).
        # The corresponding GT4Py tests are disabled (pytest marker `uses_scan_1d_field`).
        map_entry, map_exit = (None, None)
    else:
        # create map range corresponding to the field operator domain
        map_entry, map_exit = sdfg_builder.add_map(
            "fieldop",
            ctx.state,
            ndrange={
                gtir_to_sdfg_utils.get_map_variable(r.dim): f"{r.start}:{r.stop}"
                for r in field_domain
                if not sdfg_builder.is_column_axis(r.dim)
            },
        )
        assert (len(map_entry.params) + 1) == len(field_domain)

    # here we setup the edges passing through the map entry node
    for edge in input_edges:
        edge.connect(map_entry)

    # Note that `output_symbol` below is not used, we only need the tree-like
    # structure to get the type of each nested field in the `tree_map` visitor.
    dummy_output_symbol = (
        gtir_to_sdfg_utils.make_symbol_tree("__gtir_unused_dummy_var", node_type)
        if isinstance(node_type, ts.TupleType)
        else im.sym("__gtir_unused_dummy_var", node_type)
    )

    return gtx_utils.tree_map(
        lambda edge, domain, sym: (
            _create_scan_field_operator_impl(
                ctx,
                sdfg_builder,
                edge,
                domain,
                sym.type,
                map_exit,
            )
        )
    )(output, output_domain, dummy_output_symbol)


def _scan_input_name(input_name: str) -> str:
    """
    Helper function to make naming of input connectors in the scan nested SDFG
    consistent throughut this module scope.
    """
    return f"__gtir_scan_input_{input_name}"


def _scan_output_name(input_name: str) -> str:
    """
    Same as above, but for the output connecters in the scan nested SDFG.
    """
    return f"__gtir_scan_output_{input_name}"


def _lower_lambda_to_nested_sdfg(
    lambda_node: gtir.Lambda,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    field_domain: gtir_domain.FieldopDomain,
    init_data: gtir_to_sdfg_types.FieldopResult,
    lambda_params: Sequence[gtir.Sym],
    scan_forward: bool,
    scan_carry_symbol: gtir.Sym,
) -> tuple[gtir_to_sdfg.SubgraphContext, gtir_to_sdfg_types.FieldopResult]:
    """
    Helper method to lower the lambda node representing the scan stencil dataflow
    inside a separate SDFG.

    In regular field operators, where the computation of a grid point is independent
    from other points, therefore the stencil can be lowered to a mapped tasklet
    dataflow, and the map range is defined on the full domain.
    The scan field operator has to carry an intermediate result while the stencil
    is applied on vertical levels, which is input to the computation of next level
    (an accumulator function, for example). Therefore, the points on the vertical
    dimension are computed inside a `LoopRegion` construct.
    This function creates the `LoopRegion` inside a nested SDFG, which will be
    mapped by the caller to the horizontal domain in the field operator context.

    Args:
        lambda_node: The lambda representing the stencil expression on the scan dimension.
        ctx: The SDFG context where the scan field operator is translated.
        sdfg_builder: The SDFG builder object to access the field operator context.
        field_domain: The field operator domain, with all horizontal and vertical dimensions.
        init_data: The data produced in the field operator context that is used
            to initialize the scan carry value.
        lambda_params: List of symbols used as parameters of the lambda expression.
        scan_forward: When True, the loop should range starting from the origin;
            when False, traverse towards origin.
        scan_carry_symbol: The symbol used in the stencil expression to carry the
            intermediate result along the vertical dimension.

    Returns:
        A tuple of two elements:
          - The subgraph context containing the `LoopRegion` along the vertical
            dimension, to be instantied as a nested SDFG in the field operator context.
          - The inner fields, that is 1d arrays with vertical shape containing
            the output of the stencil computation. These fields will have to be
            mapped to outer arrays by the caller. The caller is responsible to ensure
            that inner and outer arrays use the same strides.
    """

    # We pass an empty set as symbolic arguments, which implies that all scalar
    # inputs of the scan nested SDFG will be represented as scalar data containers.
    # The reason why we do not check for dace symbolic expressions and do not map
    # them to inner symbols is that the scan expression should not contain any domain
    # expression (no field operator inside).
    assert not any(
        eve.walk_values(lambda_node).map(
            lambda node: cpm.is_call_to(node, ("cartesian_domain", "unstructured_domain"))
        )
    )
    # the lambda expression, i.e. body of the scan, will be created inside a nested SDFG.
    lambda_ctx = sdfg_builder.setup_nested_context(
        lambda_node,
        "scan",
        ctx,
        lambda_params,
        symbolic_inputs=set(),
        capture_scope_symbols=False,
    )

    # We set `using_explicit_control_flow=True` because the vertical scan is lowered to a `LoopRegion`.
    # This property is used by pattern matching in SDFG transformation framework
    # to skip those transformations that do not yet support control flow blocks.
    lambda_ctx.sdfg.using_explicit_control_flow = True

    # We use the entry state for initialization of the scan carry variable.
    init_state = lambda_ctx.state

    # use the vertical dimension in the domain as scan dimension
    scan_domain = next(r for r in field_domain if sdfg_builder.is_column_axis(r.dim))

    # extract the scan loop range
    scan_loop_var = gtir_to_sdfg_utils.get_map_variable(scan_domain.dim)

    # in case the scan operator computes a list (not a scalar), we need to add an extra dimension
    def get_scan_output_shape(
        scan_init_data: gtir_to_sdfg_types.FieldopData,
    ) -> list[dace.symbolic.SymExpr]:
        scan_column_size = scan_domain.stop - scan_domain.start
        if isinstance(scan_init_data.gt_type, ts.ScalarType):
            return [scan_column_size]
        assert isinstance(scan_init_data.gt_type, ts.ListType)
        assert scan_init_data.gt_type.offset_type
        offset_type = scan_init_data.gt_type.offset_type
        offset_provider_type = sdfg_builder.get_offset_provider_type(offset_type.value)
        assert isinstance(offset_provider_type, gtx_common.NeighborConnectivityType)
        list_size = offset_provider_type.max_neighbors
        return [scan_column_size, dace.symbolic.SymExpr(list_size)]

    if isinstance(init_data, tuple):
        lambda_result_shape = gtx_utils.tree_map(get_scan_output_shape)(init_data)
    else:
        assert init_data is not None
        lambda_result_shape = get_scan_output_shape(init_data)

    # Create a loop region over the vertical dimension corresponding to the scan column
    if scan_forward:
        scan_loop = dace.sdfg.state.LoopRegion(
            label="scan",
            condition_expr=f"{scan_loop_var} < {scan_domain.stop}",
            loop_var=scan_loop_var,
            initialize_expr=f"{scan_loop_var} = {scan_domain.start}",
            update_expr=f"{scan_loop_var} = {scan_loop_var} + 1",
            inverted=False,
        )
    else:
        scan_loop = dace.sdfg.state.LoopRegion(
            label="scan",
            condition_expr=f"{scan_loop_var} >= {scan_domain.start}",
            loop_var=scan_loop_var,
            initialize_expr=f"{scan_loop_var} = {scan_domain.stop} - 1",
            update_expr=f"{scan_loop_var} = {scan_loop_var} - 1",
            inverted=False,
        )
    lambda_ctx.sdfg.add_node(scan_loop, ensure_unique_name=True)
    lambda_ctx.sdfg.add_edge(init_state, scan_loop, dace.InterstateEdge())

    # Inside the loop region, create a 'compute' and an 'update' state.
    # The body of the 'compute' state implements the stencil expression for one vertical level.
    # The 'update' state writes the value computed by the stencil into the scan carry variable,
    # in order to make it available to the next vertical level.
    compute_state = scan_loop.add_state("scan_compute")
    compute_ctx = gtir_to_sdfg.SubgraphContext(
        lambda_ctx.sdfg, compute_state, lambda_ctx.scope_symbols
    )
    update_state = scan_loop.add_state_after(compute_state, "scan_update")

    # inside the 'compute' state, visit the list of arguments to be passed to the stencil
    stencil_args = [
        _parse_scan_fieldop_arg(im.ref(p.id), compute_ctx, sdfg_builder, field_domain)
        for p in lambda_node.params
    ]
    # stil inside the 'compute' state, generate the dataflow representing the stencil
    # to be applied on the horizontal domain
    lambda_input_edges, lambda_result = gtir_dataflow.translate_lambda_to_dataflow(
        compute_ctx.sdfg, compute_ctx.state, sdfg_builder, lambda_node, stencil_args
    )
    # connect the dataflow input directly to the source data nodes, without passing through a map node;
    # the reason is that the map for horizontal domain is outside the scan loop region
    for edge in lambda_input_edges:
        edge.connect(map_entry=None)

    # Create the body of the initialization state, which will initialize the scan
    # carry variable. Note that we do it here, after having visited the lambda node
    # and having built 'compute_state', for a specific reason: the scan carry variable
    # was initially a lambda parameter, and threfore lowered as a non-transient node,
    # but it needs to be changed into transient (see `scan_carry_desc.transient = True`)
    # because it is only used as internal state. The input data will a different node,
    # which is copied to the scan carry variable in the initialization state.
    scan_carry_input = (
        gtir_to_sdfg_utils.make_symbol_tree(scan_carry_symbol.id, scan_carry_symbol.type)
        if isinstance(scan_carry_symbol.type, ts.TupleType)
        else scan_carry_symbol
    )

    def init_scan_carry(sym: gtir.Sym) -> None:
        scan_carry_dataname = str(sym.id)
        scan_carry_desc = lambda_ctx.sdfg.data(scan_carry_dataname)
        input_scan_carry_dataname = _scan_input_name(scan_carry_dataname)
        input_scan_carry_desc = scan_carry_desc.clone()
        lambda_ctx.sdfg.add_datadesc(input_scan_carry_dataname, input_scan_carry_desc)
        scan_carry_desc.transient = True
        init_state.add_nedge(
            init_state.add_access(input_scan_carry_dataname),
            init_state.add_access(scan_carry_dataname),
            lambda_ctx.sdfg.make_array_memlet(input_scan_carry_dataname),
        )

    if isinstance(scan_carry_input, tuple):
        gtx_utils.tree_map(init_scan_carry)(scan_carry_input)
    else:
        init_scan_carry(scan_carry_input)

    # connect the dataflow output nodes, called 'scan_result' below, to a global field called 'output'
    output_column_index = dace.symbolic.pystr_to_symbolic(scan_loop_var) - scan_domain.start

    def connect_scan_output(
        scan_output_edge: gtir_dataflow.DataflowOutputEdge,
        scan_output_shape: list[dace.symbolic.SymExpr],
        scan_carry_sym: gtir.Sym,
    ) -> gtir_to_sdfg_types.FieldopData:
        scan_result = scan_output_edge.result
        if isinstance(scan_result.gt_dtype, ts.ScalarType):
            assert scan_result.gt_dtype == scan_carry_sym.type
            # the scan field operator computes a column of scalar values
            assert len(scan_output_shape) == 1
            output_subset = dace_subsets.Range.from_string(str(output_column_index))
        else:
            raise NotImplementedError("scan with list output is not supported.")
        scan_result_data = scan_result.dc_node.data
        scan_result_desc = scan_result.dc_node.desc(lambda_ctx.sdfg)
        scan_result_subset = dace_subsets.Range.from_array(scan_result_desc)

        # `sym` represents the global output data, that is the nested-SDFG output connector
        scan_carry_data = str(scan_carry_sym.id)
        output = _scan_output_name(scan_carry_data)
        # Note that we set `transient=True` because the lowering expects the dataflow
        # of nested SDDFG to write to some internal temporary nodes. These data elements
        # should be turned into globals by the caller and handled as output connections.
        lambda_ctx.sdfg.add_array(output, scan_output_shape, scan_result_desc.dtype, transient=True)
        output_node = compute_state.add_access(output)

        # in the 'compute' state, we write the current vertical level data to the output field
        # (the output field is mapped to an external array)
        compute_state.add_nedge(
            scan_result.dc_node,
            output_node,
            dace.Memlet(data=output, subset=output_subset, other_subset=scan_result_subset),
        )

        # in the 'update' state, the value of the current vertical level is written
        # to the scan carry variable for the next loop iteration
        update_state.add_nedge(
            update_state.add_access(scan_result_data),
            update_state.add_access(scan_carry_data),
            dace.Memlet(
                data=scan_result_data, subset=scan_result_subset, other_subset=scan_result_subset
            ),
        )

        output_type = ts.FieldType(dims=[scan_domain.dim], dtype=scan_result.gt_dtype)
        return gtir_to_sdfg_types.FieldopData(output_node, output_type, origin=(scan_domain.start,))

    # write the stencil result (value on one vertical level) into a 1D field
    # with full vertical shape representing one column
    lambda_output = gtx_utils.tree_map(connect_scan_output)(
        lambda_result, lambda_result_shape, scan_carry_input
    )

    return lambda_ctx, lambda_output


def _handle_dataflow_result_of_nested_sdfg(
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    nsdfg_node: dace_nodes.NestedSDFG,
    inner_ctx: gtir_to_sdfg.SubgraphContext,
    outer_ctx: gtir_to_sdfg.SubgraphContext,
    inner_data: gtir_to_sdfg_types.FieldopData,
    output_domain: infer_domain.NonTupleDomainAccess,
) -> gtir_dataflow.DataflowOutputEdge | None:
    assert isinstance(inner_data.gt_type, ts.FieldType)
    inner_dataname = inner_data.dc_node.data
    inner_desc = inner_ctx.sdfg.data(inner_dataname)
    assert inner_desc.transient

    if isinstance(output_domain, infer_domain.DomainAccessDescriptor):
        # The field is not used outside the nested SDFG. It is likely just storage
        # for some internal state, accessed during column scan, and can be turned
        # into a transient array inside the nested SDFG.
        assert output_domain == infer_domain.DomainAccessDescriptor.NEVER
        nsdfg_node.out_connectors.pop(inner_dataname)
        return None

    # The field is used outside the nested SDFG, therefore it needs to be copied
    # to a temporary array in the parent SDFG (outer context).
    field_dims, _, field_shape = gtir_domain.get_field_layout(
        gtir_domain.get_field_domain(output_domain)
    )
    scan_dim_index = [sdfg_builder.is_column_axis(dim) for dim in field_dims].index(True)
    scan_column_size = field_shape[scan_dim_index]

    if bool(scan_column_size == 1):
        # Special case where we only write the last level of the scan column.
        # We represent the inner result as a single value and let the scan loop override it.
        inner_desc = dace.data.Scalar(inner_desc.dtype, transient=False)
        inner_ctx.sdfg.arrays[inner_dataname] = inner_desc
        # We write the result to the output field after the loop region has ended.
        if len(inner_ctx.sdfg.states()) == 3:
            assert sorted(st.label for st in inner_ctx.sdfg.states()) == [
                "scan_compute",
                "scan_entry",
                "scan_update",
            ]
            scan_loop = next(
                node
                for node in inner_ctx.sdfg.nodes()
                if isinstance(node, dace.sdfg.state.LoopRegion)
            )
            last_level_state = inner_ctx.sdfg.add_state_after(scan_loop, "scan_last_level")
        else:
            assert sorted(st.label for st in inner_ctx.sdfg.states()) == [
                "scan_compute",
                "scan_entry",
                "scan_last_level",
                "scan_update",
            ]
            last_level_state = next(
                s for s in inner_ctx.sdfg.states() if s.label == "scan_last_level"
            )
        # Update the write edge inside the scan nested SDFG to write into a scalar rather than a 1D array.
        scan_compute_state = next(s for s in inner_ctx.sdfg.states() if s.label == "scan_compute")
        inner_output_node = next(
            n for n in scan_compute_state.data_nodes() if n.data == inner_dataname
        )
        assert (
            scan_compute_state.in_degree(inner_output_node) == 1
            and scan_compute_state.out_degree(inner_output_node) == 0
        )
        inner_write_edge = scan_compute_state.in_edges(inner_output_node)[0]
        assert isinstance(inner_write_edge.src, dace_nodes.AccessNode)
        assert isinstance(inner_write_edge.src.desc(inner_ctx.sdfg), dace.data.Scalar)
        scan_compute_state.remove_node(inner_output_node)
        last_level_state.add_nedge(
            last_level_state.add_access(inner_write_edge.src.data),
            last_level_state.add_access(inner_dataname),
            dace.Memlet(data=inner_dataname, subset="0", other_subset="0"),
        )
    else:
        inner_desc.transient = False

    outer_dataname, outer_desc = sdfg_builder.add_temp_array(
        outer_ctx.sdfg, (scan_column_size,), inner_desc.dtype
    )
    outer_node = outer_ctx.state.add_access(outer_dataname)
    outer_ctx.state.add_edge(
        nsdfg_node,
        inner_dataname,
        outer_node,
        None,
        dace.Memlet.from_array(outer_dataname, outer_desc),
    )
    output_expr = gtir_dataflow.ValueExpr(outer_node, inner_data.gt_type.dtype)
    return gtir_dataflow.DataflowOutputEdge(outer_ctx.state, output_expr)


def translate_scan(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Generates the dataflow subgraph for the `as_fieldop` builtin with a scan operator.

    It differs from `translate_as_fieldop()` in that the horizontal domain is lowered
    to a map scope, while the scan column computation is lowered to a `LoopRegion`
    on the vertical dimension, that is inside the horizontal map.
    The current design choice is to keep the map scope on the outer level, and
    the `LoopRegion` inside. This choice follows the GTIR representation where
    the `scan` operator is called inside the `as_fieldop` node.

    Implements the `PrimitiveTranslator` protocol.
    """
    assert isinstance(node, gtir.FunCall)
    assert cpm.is_call_to(node.fun, "as_fieldop")
    assert isinstance(node.type, (ts.FieldType, ts.TupleType))

    fun_node = node.fun
    assert len(fun_node.args) == 2
    scan_expr, scan_domain_expr = fun_node.args
    assert cpm.is_call_to(scan_expr, "scan")

    # parse the domain of the scan field operator
    assert isinstance(scan_domain_expr.type, ts.DomainType)
    field_domain = gtir_domain.get_field_domain(
        domain_utils.SymbolicDomain.from_expr(scan_domain_expr)
    )

    # parse scan parameters
    assert len(scan_expr.args) == 3
    stencil_expr = scan_expr.args[0]
    assert isinstance(stencil_expr, gtir.Lambda)

    # params[0]: the lambda parameter to propagate the scan carry on the vertical dimension
    scan_carry = stencil_expr.params[0].id
    scan_carry_type = stencil_expr.params[0].type
    assert isinstance(scan_carry_type, ts.DataType)

    # params[1]: boolean flag for forward/backward scan
    assert isinstance(scan_expr.args[1], gtir.Literal) and ti.is_logical(scan_expr.args[1].type)
    scan_forward = scan_expr.args[1].value == "True"

    # params[2]: the expression that computes the value for scan initialization
    init_expr = scan_expr.args[2]
    # visit the initialization value of the scan expression
    init_data = sdfg_builder.visit(init_expr, ctx=ctx)

    # define the symbols passed as parameter to the lambda expression, which consists
    # of the carry argument and all lambda function arguments
    lambda_arg_types: list[ts.DataType] = [scan_carry_type] + [
        arg.type for arg in node.args if isinstance(arg.type, ts.DataType)
    ]
    lambda_params = [
        im.sym(p.id, arg_type)
        for p, arg_type in zip(stencil_expr.params, lambda_arg_types, strict=True)
    ]

    # lower the scan stencil expression in a separate SDFG context
    lambda_ctx, lambda_output = _lower_lambda_to_nested_sdfg(
        stencil_expr,
        ctx,
        sdfg_builder,
        field_domain,
        init_data,
        lambda_params,
        scan_forward,
        im.sym(scan_carry, scan_carry_type),
    )

    # visit the arguments to be passed to the lambda expression
    # this must be executed before visiting the lambda expression, in order to populate
    # the data descriptor with the correct field domain offsets for field arguments
    lambda_args = [sdfg_builder.visit(arg, ctx=ctx) for arg in node.args]
    lambda_args_mapping = [
        (im.sym(_scan_input_name(scan_carry), scan_carry_type), init_data),
    ] + [
        (gt_symbol, arg)
        for gt_symbol, arg in zip(stencil_expr.params[1:], lambda_args, strict=True)
    ]

    lambda_arg_nodes, symbolic_args = gtir_to_sdfg.flatten_tuple_args(lambda_args_mapping)

    # The lambda expression of a scan field operator should never capture symbols
    # from the ouside scope, therefore we call `add_nested_sdfg()` with `capture_outer_data=False`.
    nsdfg_node, input_memlets = sdfg_builder.add_nested_sdfg(
        node=stencil_expr,
        inner_ctx=lambda_ctx,
        outer_ctx=ctx,
        symbolic_args=symbolic_args,
        data_args=lambda_arg_nodes,
        inner_result=lambda_output,
        capture_outer_data=False,
    )

    # Block the inlining of NestedSDFG containing a scan.
    #  We do this to avoid a bug in DaCe simplify transformations, see
    #  [issue#2182](https://github.com/spcl/dace/issues/2182) for more. Before the bug
    #  was hidden but once we started running `TrivialTaskletElimination` it was no
    #  longer the case. The solution is to block the inlining and keep scans localized
    #  inside their NestedSDFG.
    # NOTE: Currently there is no transformation that takes advantages in any way
    #   of a scan and they are mostly inside Maps anyway, except of unit tests,
    #   where inlining is not possible anyway.
    nsdfg_node.no_inline = True

    input_edges = []
    for input_connector, memlet in input_memlets.items():
        src_node = lambda_arg_nodes[input_connector]
        assert src_node is not None
        input_edge = gtir_dataflow.MemletInputEdge(
            ctx.state, src_node.dc_node, memlet.src_subset, nsdfg_node, input_connector
        )
        input_edges.append(input_edge)

    # for output connections, we create temporary arrays that contain the computation
    # results of a column slice for each point in the horizontal domain
    output_tree = gtx_utils.tree_map(
        lambda output_data, output_domain: _handle_dataflow_result_of_nested_sdfg(
            sdfg_builder=sdfg_builder,
            nsdfg_node=nsdfg_node,
            inner_ctx=lambda_ctx,
            outer_ctx=ctx,
            inner_data=output_data,
            output_domain=output_domain,
        )
    )(lambda_output, node.annex.domain)

    # we call a helper method to create a map scope that will compute the entire field
    return _create_scan_field_operator(
        ctx, field_domain, node.type, sdfg_builder, input_edges, output_tree, node.annex.domain
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.iterator.transforms.pass_manager import (
    GTIRTransform,
    apply_common_transforms,
    apply_fieldview_transforms,
)


__all__ = ["GTIRTransform", "apply_common_transforms", "apply_fieldview_transforms"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
from typing import Any, Optional

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes
from dace.transformation import helpers

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dace_properties.make_properties
class RemovePointwiseViews(dace_transformation.SingleStateTransformation):
    """
    Remove pointwise views from the SDFG.
    This transformation is used to remove views that are created for pointwise operations.
    It redirects the edges from the view to the original data and removes the view node.
    The view generation is non-deterministic and usually happens after reduction `Library` nodes.
    """

    source_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    destination_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = None
        if single_use_data is not None:
            self._single_use_data = single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.source_node, cls.destination_node)]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        src_node: dace_nodes.AccessNode = self.source_node
        dst_node: dace_nodes.AccessNode = self.destination_node
        src_desc: dace_data.Data = src_node.desc(sdfg)
        dst_desc: dace_data.Data = dst_node.desc(sdfg)

        scope = graph.scope_dict()

        # Currently handle only pointwise views inside a Map. If the view is not inside a Map,
        # it is probably not a pointwise view.
        if scope[src_node] is None or scope[dst_node] is None:
            return False  # Not inside a Map.

        if not src_desc.transient:
            return False
        if not gtx_transformations.utils.is_view(src_desc, sdfg):
            return False
        if graph.in_degree(src_node) > 1:
            return False
        if not dst_desc.transient:
            return False
        if graph.out_degree(src_node) != 1:
            return False
        if len(graph.in_edges(src_node)) != 1:
            return False
        if isinstance(src_desc, dace_data.ArrayView) and src_desc.total_size != 1:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        src_node: dace_nodes.AccessNode = self.source_node
        dst_node: dace_nodes.AccessNode = self.destination_node

        dst_in_edge = graph.in_edges(dst_node)[0]
        edge_to_redirect = graph.in_edges(src_node)[0]
        # Redirect the edge going into the view to the AccessNode that is the destination of the view.
        # new_dst: AccessNode that is the destination of the view.
        # new_dst_conn: Connection name of the destination AccessNode.
        # new_edge.data: The original Memlet of the edge_to_redirect.
        dst_in_edge_dst_subset = copy.deepcopy(dst_in_edge.data.get_dst_subset(dst_in_edge, graph))
        new_edge = helpers.redirect_edge(
            graph, edge_to_redirect, new_dst=dst_in_edge.dst, new_dst_conn=dst_in_edge.dst_conn
        )

        # If new_edge.data has the same data as the src_node, we need to update it to point to the dst_node
        # since the src_node is being removed. Else the new_edge.data can either be some other unrelated `data`
        # or dst_node.data. In this case we have to update the `other_subset` to match the dst_node.data other_subset
        # which is the subset which gets written to the dst of the edge.
        if new_edge.data.data == src_node.data:
            new_edge.data.data = dst_node.data
            new_edge.data.subset = dst_in_edge_dst_subset
        else:
            new_edge.data.other_subset = dst_in_edge_dst_subset

        graph.remove_edge(dst_in_edge)
        sdfg.remove_data(src_node.data, validate=False)
        graph.remove_node(src_node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Any, Callable

import numpy as np
import sympy

from gt4py.eve import codegen
from gt4py.eve.codegen import FormatTemplate as as_fmt
from gt4py.next.iterator import builtins, ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm


MATH_BUILTINS_MAPPING = {
    "abs": "abs({})",
    "neg": "(- {})",
    "sin": "math.sin({})",
    "cos": "math.cos({})",
    "tan": "math.tan({})",
    "arcsin": "asin({})",
    "arccos": "acos({})",
    "arctan": "atan({})",
    "sinh": "math.sinh({})",
    "cosh": "math.cosh({})",
    "tanh": "math.tanh({})",
    "arcsinh": "asinh({})",
    "arccosh": "acosh({})",
    "arctanh": "atanh({})",
    "sqrt": "math.sqrt({})",
    "exp": "math.exp({})",
    "log": "math.log({})",
    "gamma": "tgamma({})",
    "cbrt": "cbrt({})",
    "isfinite": "isfinite({})",
    "isinf": "isinf({})",
    "isnan": "isnan({})",
    "floor": "math.ifloor({})",
    "ceil": "ceil({})",
    "trunc": "trunc({})",
    "minimum": "min({}, {})",
    "maximum": "max({}, {})",
    "fmod": "fmod({}, {})",
    "power": "math.pow({}, {})",
    "float": "dace.float64({})",
    "float32": "dace.float32({})",
    "float64": "dace.float64({})",
    "int": "dace.int32({})" if np.dtype(int).itemsize == 4 else "dace.int64({})",
    "int32": "dace.int32({})",
    "int64": "dace.int64({})",
    "bool": "dace.bool_({})",
    "plus": "({} + {})",
    "minus": "({} - {})",
    "multiplies": "({} * {})",
    "divides": "({} / {})",
    "floordiv": "({} // {})",
    "eq": "({} == {})",
    "not_eq": "({} != {})",
    "less": "({} < {})",
    "less_equal": "({} <= {})",
    "greater": "({} > {})",
    "greater_equal": "({} >= {})",
    "and_": "({} and {})",
    "or_": "({} or {})",
    "xor_": "({} != {})",
    "mod": "({} % {})",
    "not_": "(not {})",
}


def _builtin_cast(val: str, target_type: str) -> str:
    assert target_type in builtins.TYPE_BUILTINS
    return MATH_BUILTINS_MAPPING[target_type].format(val)


def _builtin_get_domain_range(field: str, axis: str) -> str:
    # The builtin function returns a tuple of two values, the range start and stop.
    # Here we return part of the symbol name: the full name also contains an
    # additional suffix '_0' for range start or '_1' for stop, which is obtained
    # from to the `tuple_get` index.
    return f"__{field}_{axis}_range"


def _builtin_if(cond: str, true_val: str, false_val: str) -> str:
    return f"{true_val} if {cond} else {false_val}"


def _builtin_tuple_get(index: str, tuple_name: str) -> str:
    return f"{tuple_name}_{index}"


def _builtin_make_const_list(arg: str) -> str:
    """
    Takes a single scalar argument and broadcasts this value on the local dimension
    of map expression. In a dataflow, we represent it as a tasklet that writes
    a value to a scalar node.
    """
    return arg


GENERAL_BUILTIN_MAPPING: dict[str, Callable[..., str]] = {
    "cast_": _builtin_cast,
    "get_domain_range": _builtin_get_domain_range,
    "if_": _builtin_if,
    "make_const_list": _builtin_make_const_list,
    "tuple_get": _builtin_tuple_get,
}


def format_builtin(builtin: str, *args: Any) -> str:
    if builtin in MATH_BUILTINS_MAPPING:
        fmt = MATH_BUILTINS_MAPPING[builtin]
        return fmt.format(*args)
    elif builtin in GENERAL_BUILTIN_MAPPING:
        expr_func = GENERAL_BUILTIN_MAPPING[builtin]
        return expr_func(*args)
    else:
        raise NotImplementedError(f"'{builtin}' not implemented.")


class PythonCodegen(codegen.TemplatedGenerator):
    """Helper class to visit a symbolic expression and translate it to Python code.

    The generated Python code can be use either as the body of a tasklet node or,
    as in the case of field domain definitions, for sybolic array shape and map range.
    """

    Literal = as_fmt("{value}")

    def visit_AxisLiteral(self, node: gtir.AxisLiteral, **kwargs: Any) -> str:
        return node.value

    def visit_FunCall(self, node: gtir.FunCall, args_map: dict[str, gtir.Node]) -> str:
        if isinstance(node.fun, gtir.Lambda):
            # update the mapping from lambda parameters to corresponding argument expressions
            lambda_args_map = args_map | {
                p.id: arg for p, arg in zip(node.fun.params, node.args, strict=True)
            }
            return self.visit(node.fun.expr, args_map=lambda_args_map)
        elif cpm.is_call_to(node, "deref"):
            assert len(node.args) == 1
            if not isinstance(node.args[0], gtir.SymRef):
                # shift expressions are not expected in this visitor context
                raise NotImplementedError(f"Unexpected deref with arg type '{type(node.args[0])}'.")
            return self.visit(node.args[0], args_map=args_map)
        elif isinstance(node.fun, gtir.SymRef):
            args = self.visit(node.args, args_map=args_map)
            builtin_name = str(node.fun.id)
            return format_builtin(builtin_name, *args)
        raise NotImplementedError(f"Unexpected 'FunCall' node ({node}).")

    def visit_InfinityLiteral(self, node: gtir.InfinityLiteral, **kwargs: Any) -> str:
        return str(sympy.oo) if node == gtir.InfinityLiteral.POSITIVE else str(-sympy.oo)

    def visit_SymRef(self, node: gtir.SymRef, args_map: dict[str, gtir.Node]) -> str:
        symbol = str(node.id)
        if symbol in args_map:
            return self.visit(args_map[symbol], args_map=args_map)
        return symbol


def get_source(node: gtir.Node) -> str:
    """
    Specialized visit method for symbolic expressions.

    The visitor uses `args_map` to map lambda parameters to the corresponding argument expressions.

    Returns:
        A string containing the Python code corresponding to a symbolic expression
    """
    return PythonCodegen.apply(node, args_map={})
